# coding=utf-8

import datetime as dt

from flask import current_app, request
from jwt import InvalidTokenError

from guniflask.utils.security import encode_jwt, decode_jwt
from guniflask.security.user import AnonymousUser, User
from guniflask.security.auth import AuthManager

__all__ = ['JwtManager']


class JwtManager(AuthManager):
    def __init__(self, app=None):
        super().__init__(app=app)
        self._user_loader = self._load_user_from_header

    def init_app(self, app):
        self._set_default_app_config(app)
        app.extensions['jwt_manager'] = self
        super().init_app(app)

    def _set_default_app_config(self, app):
        app.config.setdefault('JWT_ALGORITHM', 'HS256')
        app.config.setdefault('JWT_SECRET', None)
        app.config.setdefault('JWT_PRIVATE_KEY', None)
        app.config.setdefault('JWT_PUBLIC_KEY', None)
        app.config.setdefault('JWT_ACCESS_TOKEN_EXPIRES_IN', dt.timedelta(days=1))

    def create_access_token(self, authorities=None, expires_in=None, **kwargs):
        config = current_app.config
        user_claims = dict(kwargs)
        user_claims['authorities'] = authorities
        payload = {'user_claims': user_claims}
        if expires_in is None:
            expires_in = config['JWT_ACCESS_TOKEN_EXPIRES_IN']
        return encode_jwt(payload, config['JWT_SECRET'] or config['JWT_PRIVATE_KEY'],
                          config['JWT_ALGORITHM'], expires_in=expires_in)

    def _load_user_from_header(self):
        config = current_app.config
        user = AnonymousUser()
        token = self.current_token
        if token:
            try:
                payload = decode_jwt(token, config['JWT_SECRET'] or config['JWT_PUBLIC_KEY'],
                                     config['JWT_ALGORITHM'])
            except InvalidTokenError:
                pass
            else:
                user_claims = payload['user_claims']
                user = User(**user_claims)
        return user

    @property
    def current_token(self):
        token = self._get_token_from_header()
        if token is None:
            token = self._get_token_from_query()
        return token

    @staticmethod
    def _get_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.startswith('Bearer'):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _get_token_from_query():
        return request.args.get('access_token')



# coding=utf-8

__all__ = ['blueprint', 'route', 'get_route', 'post_route', 'put_route', 'patch_route', 'delete_route']


def blueprint(**options):
    def decorator(func):
        func.__is_blueprint = True
        func.__options = options
        return func

    return decorator


def route(rule, **options):
    def decorator(func):
        func.__is_route = True
        func.__rule = rule
        func.__options = options
        return func

    return decorator


def get_route(rule, **options):
    options['methods'] = ['GET']
    return route(rule, **options)


def post_route(rule, **options):
    options['methods'] = ['POST']
    return route(rule, **options)


def put_route(rule, **options):
    options['methods'] = ['PUT']
    return route(rule, **options)


def patch_route(rule, **options):
    options['methods'] = ['PATCH']
    return route(rule, **options)


def delete_route(rule, **options):
    options['methods'] = ['DELETE']
    return route(rule, **options)



# coding=utf-8


__all__ = ['User', 'AnonymousUser']


class User:
    ROLE_PREFIX = 'role_'

    def __init__(self, authorities=None, **kwargs):
        self.authorities = []
        if authorities is not None:
            for a in authorities:
                self.authorities.append(a.lower())
        self._authorities_set = set(self.authorities)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def is_authenticated(self):
        return True

    def has_authority(self, authority):
        authority = authority.lower()
        return authority in self._authorities_set

    def has_any_authority(self, *authorities):
        for a in authorities:
            if self.has_authority(a):
                return True
        return False

    def has_role(self, role):
        return self.has_authority(self._get_role_with_prefix(self.ROLE_PREFIX, role))

    def has_any_role(self, *roles):
        for r in roles:
            if self.has_role(r):
                return True
        return False

    @staticmethod
    def _get_role_with_prefix(prefix, role):
        if role is None or prefix is None:
            return role
        if role.startswith(prefix):
            return role
        return prefix + role


class AnonymousUser(User):
    def __init__(self, authorities=None, **kwargs):
        super().__init__(**kwargs)

    @property
    def is_authenticated(self):
        return False



# coding=utf-8

import os
from multiprocessing import Process
import signal
from threading import Thread
import time
from os.path import join
from logging.config import dictConfig
import logging

from gunicorn.glogging import Logger

from guniflask.app import create_bg_process_app
from guniflask.utils.config import load_object


class BgProcess:
    def __init__(self, app):
        self.app = app
        self.settings = app.extensions['settings']
        self.configure_logging()
        self.logger = logging.getLogger(self.app.name + '.bg')

    def configure_logging(self):
        if 'bg_log_config' in self.settings:
            dictConfig(self.settings['bg_log_config'])
        else:
            logger = logging.getLogger(self.app.name)
            debug = os.environ.get('GUNIFLASK_DEBUG')
            if debug:
                logger.setLevel(logging.DEBUG)
            else:
                logger.setLevel(logging.INFO)
            log_dir = os.environ['GUNIFLASK_LOG_DIR']
            id_string = os.environ['GUNIFLASK_ID_STRING']
            project_name = os.environ.get('GUNIFLASK_PROJECT_NAME')
            log_file = join(log_dir, 'bg-{}-{}.log'.format(project_name, id_string))
            handler = logging.FileHandler(log_file)
            formatter = logging.Formatter(Logger.error_fmt, Logger.datefmt)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

    def run(self):
        raise NotImplemented

    def start(self):
        try:
            with self.app.app_context():
                self.run()
        except Exception:
            self.logger.error('error occurred in %s', self.__class__.__name__, exc_info=True)


class BgProcessRunner:
    def __init__(self, name=None, bg_cls=None, on_starting=None):
        self.name = name
        self.bg_cls = bg_cls
        self.on_starting = on_starting
        self.debug = os.environ.get('GUNIFLASK_DEBUG')
        self.bg = None

    def start(self, server):
        if self.on_starting is not None:
            self.on_starting(server)
        pid = os.getpid()
        p = Process(target=self._start_bg_process, args=(pid,))
        p.start()
        p.join()

    def _start_bg_process(self, pid):
        daemonize()
        self._bg_init_signals()
        t = Thread(target=self._bg_supervisor, args=(pid,))
        t.start()
        app = create_bg_process_app(self.name)
        bg_cls = self.bg_cls
        if isinstance(bg_cls, str):
            bg_cls = load_object(bg_cls)
        self.bg = bg_cls(app)
        self.bg.start()

    def _bg_init_signals(self):
        def _exit(signum, frame):
            self._bg_exit()

        signal.signal(signal.SIGINT, _exit)
        signal.signal(signal.SIGTERM, _exit)

    def _bg_supervisor(self, pid):
        while True:
            if not existsp(pid):
                self._bg_exit()
            time.sleep(1)

    def _bg_exit(self):
        if self.bg:
            self.bg.logger.info('%s exiting (pid: %s)', self.bg.__class__.__name__, os.getpid())
        os._exit(0)


def daemonize():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def existsp(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True



# coding=utf-8

import logging
from importlib import import_module
import inspect

from flask import Flask, Blueprint
from flask_cors import CORS
from flask_sqlalchemy import SQLAlchemy

from guniflask.config import Config
from guniflask.utils.logging import redirect_app_logger, redirect_logger
from guniflask.model import wrap_model
from guniflask.apidoc import ApiDoc
from guniflask.security import JwtManager
from guniflask.utils.config import walk_modules

log = logging.getLogger(__name__)

app_default_settings = {
    'debug': False,
    'cors': True,
    'jwt': False,
    # Flask-SQLAlchemy
    'SQLALCHEMY_TRACK_MODIFICATIONS': False
}

sqlalchemy_instances = []


def new_sqlalchemy(*args, **kwargs) -> SQLAlchemy:
    obj = SQLAlchemy(*args, **kwargs)
    sqlalchemy_instances.append(obj)
    return obj


def _get_app_module(app):
    return import_module(app.name + '.app')


def _get_instance_from_app(app, name):
    app_module = _get_app_module(app)
    obj = getattr(app_module, name, None)
    return obj


def set_app_config(app):
    config = Config()
    config.init_app(app)
    s = config.app_settings(app)
    for k, v in app_default_settings.items():
        s.setdefault(k, v)

    app_module = _get_app_module(app)
    _make_settings = getattr(app_module, 'make_settings', None)
    if _make_settings:
        _make_settings(app, s)

    for k, v in s.items():
        if k.isupper():
            app.config[k] = v

    log.info('%s active profiles: %s', app.name, s['active_profiles'])


def init_app(app):
    s = app.extensions['settings']

    # CORS
    if s['cors']:
        if isinstance(s['cors'], dict):
            CORS(app, **s['cors'])
        else:
            CORS(app)

    # database configuration
    for db in sqlalchemy_instances:
        wrap_model(db.Model)

    # authentication
    if s['jwt']:
        jwt_manager = JwtManager()
        jwt_manager.init_app(app)

    # API doc
    if s['apidoc'] or (s['apidoc'] is None and s['debug']):
        apidoc = ApiDoc()
        apidoc.init_app(app)

    app_module = _get_app_module(app)
    _init_app = getattr(app_module, 'init_app', None)
    if _init_app:
        _init_app(app, s)


def create_app(name):
    app = Flask(name)
    gunicorn_logger = logging.getLogger('gunicorn.error')
    redirect_app_logger(app, gunicorn_logger)
    redirect_logger(name, gunicorn_logger)
    set_app_config(app)
    init_app(app)
    with app.app_context():
        register_blueprints(app)
    return app


def register_blueprints(app):
    def iter_blueprints():
        for module in walk_modules(app.name):
            for obj in vars(module).values():
                if isinstance(obj, Blueprint):
                    yield obj
                if inspect.isclass(obj) and hasattr(obj, '__is_blueprint'):
                    b = Blueprint(obj.__name__, obj.__module__, **getattr(obj, '__options'))
                    o = obj()
                    for k in dir(o):
                        f = getattr(o, k)
                        if hasattr(f, '__is_route'):
                            b.add_url_rule(getattr(f, '__rule'), endpoint=f.__name__, view_func=f,
                                           **getattr(f, '__options'))
                    yield b

    for b in iter_blueprints():
        app.register_blueprint(b)


def create_bg_process_app(name):
    app = Flask(name)
    set_app_config(app)
    init_app(app)
    return app



# coding=utf-8

from functools import wraps

from flask import current_app, _request_ctx_stack, abort
from werkzeug.local import LocalProxy

__all__ = ['current_user', 'auth_manager', 'AuthManager',
           'login_required', 'roles_required', 'authorities_required']

auth_manager = LocalProxy(lambda: current_app.extensions['auth_manager'])


def _load_user():
    ctx = _request_ctx_stack.top
    if ctx is not None:
        if not hasattr(ctx, 'user'):
            user = auth_manager.load_user()
            ctx.user = user
        return ctx.user


current_user = LocalProxy(_load_user)


class AuthManager:
    def __init__(self, app=None):
        self._user_loader = None
        self._unauthorized_handler = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.extensions['auth_manager'] = self

    def user_loader(self, func):
        self._user_loader = func
        return func

    def load_user(self):
        if self._user_loader is None:
            raise RuntimeError('Missing user loader')
        return self._user_loader()

    @property
    def unauthorized_view(self):
        if self._unauthorized_handler:
            return self._unauthorized_handler()
        return abort(401)

    def unauthorized(self, func):
        self._unauthorized_handler = func
        return func


def login_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not current_user.is_authenticated:
            return auth_manager.unauthorized_view
        return func(*args, **kwargs)

    return wrapper


def roles_required(*roles):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if current_user is None or not current_user.has_any_role(*roles):
                return auth_manager.unauthorized_view
            return func(*args, **kwargs)

        return wrapper

    return decorator


def authorities_required(*authorities):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if current_user is None or not current_user.has_any_authority(*authorities):
                return auth_manager.unauthorized_view
            return func(*args, **kwargs)

        return wrapper

    return decorator



# coding=utf-8


class UsageError(Exception):
    """
    Usage error.
    """

    def __init__(self, *args, print_help=False, **kwargs):
        self.print_help = print_help
        super().__init__(*args, **kwargs)


class AbortedError(Exception):
    """
    Aborted error.
    """


class TemplateError(Exception):
    """
    Template error.
    """



# coding=utf-8

import json
from os.path import join
from functools import wraps

from flask import request, abort, jsonify, render_template, current_app

from guniflask.utils.template import template_folder
from guniflask.config import settings
from guniflask.security import roles_required
from guniflask.web import blueprint, get_route, route

static_folder = join(template_folder, 'static')


def debug_only(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if settings['debug']:
            return func(*args, **kwargs)
        return render_template('debug_only.html'), 403

    return wrapper


accounts = {
    'root': {
        'password': '123456',
        'authorities': ['role_admin', 'role_user']
    }
}


@blueprint(url_prefix='/hello-world',
           template_folder=join(template_folder, 'hello_world'),
           static_folder=static_folder,
           static_url_path='/static')
class HelloWorld:
    """
    An example
    """

    def __init__(self):
        self.jwt_manager = current_app.extensions['jwt_manager']

    @get_route('/')
    def home_page(self):
        """
        Home page
        """
        return render_template('index.html')

    @route('/login', methods=['GET', 'POST'])
    def login(self):
        """
        Login app
        """
        if request.method == 'GET':
            return render_template('login.html')
        data = request.get_json()
        if data is not None:
            username = data.get('username')
            password = data.get('password')
            if username in accounts and password == accounts[username]['password']:
                token = self.jwt_manager.create_access_token(accounts[username]['authorities'],
                                                             username=username)
                return jsonify({'access_token': token,
                                'username': username})
        return abort(403)

    @get_route('/settings')
    @debug_only
    def get_settings(self):
        """
        Settings page
        """
        return render_template('settings.html')

    @get_route('/settings-table')
    @roles_required('admin')
    def get_settings_table(self):
        """
        Get settings of app
        """
        s = {}
        for k, v in settings.items():
            if is_jsonable(v):
                s[k] = v
        app_settings = [{'key': i, 'value': j} for i, j in s.items()]
        app_settings.sort(key=lambda k: k['key'])
        return render_template('settings_table.html', app_settings=app_settings)


def is_jsonable(v):
    try:
        json.dumps(v)
    except Exception:
        return False
    return True
